--[============================[
  =         The Client         =
--]============================]


require "constants"

require "libs.strict"
require "libs.table"
require "libs.print_r"

require "util.class"
require "util.mathutil"
require "util.vector2"
require "util.matrix"
require "util.color"

require "tuning"
require "camera"
require "ball"


global("RandomGenerator")
RandomGenerator = love.math.newRandomGenerator()

global("Client")
Client = {}
Client.width, Client.height = love.graphics.getDimensions()
Client.controls = require('controls')
Client.camera = Camera(Vector2(0,0),0,1)

global("Balls")
Balls = require('balls')

Client.load = function ()
	Client.gravity_enabled = TUNING.GRAVITY_ENABLED
	Client.gravity = Vector2(0,TUNING.WORLD.GRAVITY)

	Balls:reset()
end


Client.update = function (dt)
	Client.handleInputs()
	Client.updateCamera(dt)
	Client.updateWorld(dt)
	Client.updateBalls(dt)
end


Client.draw = function ()
	Client.camera:attach()
	Balls:draw()
	Client.camera:detatch()

	love.graphics.setColor(255,255,255)
	local debug = string.format("FPS: %s\nCamera: %s\nMouse: %s\nBalls: %s",love.timer.getFPS(),tostring(Client.camera),tostring(Client.camera:mousePos()), #Balls)
	love.graphics.print(debug)
end


Client.handleInputs = function ()
	for k,action in pairs(Client.controls) do
		action.wasDown = action.isDown
		action.isDown = false
		if action.KeyboardKey then
			if type(action.KeyboardKey) == 'table' then
				--- multiple bindings for this control
				for i,v in ipairs(action.KeyboardKey) do
					if love.keyboard.isDown(v) then
						action.isDown = true
						break
					end
				end
			else
				if love.keyboard.isDown(action.KeyboardKey) then
					action.isDown = true
				end
			end
		elseif action.GamepadButton then
			--- handle gamepad info
		end
	end
end


Client.updateCamera = function (dt)
	--- Update Camera with Inputs

	--- Update Camera Zoom
	local cur_zoom = 1 + TUNING.CAMERA.ZOOM_SPEED_MAX*dt
	if Client.controls.CameraZoomIn.isDown then
		Client.camera:zoom(cur_zoom)
	end
	if Client.controls.CameraZoomOut.isDown then
		Client.camera:zoom(1 / cur_zoom)
	end

	--- Update Angle
	local cur_rot = TUNING.CAMERA.ROTATE_SPEED_MAX*dt
	if Client.controls.CameraRotateClockwise.isDown then
		Client.camera:rotate(cur_rot)
	end
	if Client.controls.CameraRotateCounterclockwise.isDown then
		Client.camera:rotate(-cur_rot)
	end

	--- Update Camera Position
	local cur_pan_speed = (TUNING.CAMERA.PAN_SPEED_MAX / Client.camera.scale)*dt
	local offset = Vector2(0,0)
	if Client.controls.CameraPanUp.isDown then
		offset.y = offset.y - cur_pan_speed
	end
	if Client.controls.CameraPanDown.isDown then
		offset.y = offset.y + cur_pan_speed
	end
	if Client.controls.CameraPanLeft.isDown then
		offset.x = offset.x - cur_pan_speed
	end
	if Client.controls.CameraPanRight.isDown then
		offset.x = offset.x + cur_pan_speed
	end
	offset:rotate(Client.camera.rot)
	Client.camera:move(offset)
end


Client.updateWorld = function (dt)
	-- Update Gravity
	if Client.controls.GravityToggle.isDown and not Client.controls.GravityToggle.wasDown then
		Client.gravity_enabled = not Client.gravity_enabled
	end
	local gravity_x = 0
	local gravity_y = 0


	if Client.controls.GravityLeft.isDown and not Client.controls.GravityLeft.wasDown then
		gravity_x = gravity_x - 1
	end
	if Client.controls.GravityRight.isDown and not Client.controls.GravityRight.wasDown then
		gravity_x = gravity_x + 1
	end

	if Client.controls.GravityUp.isDown and not Client.controls.GravityUp.wasDown then
		gravity_y = gravity_y - 1
	end
	if Client.controls.GravityDown.isDown and not Client.controls.GravityDown.wasDown then
		gravity_y = gravity_y + 1
	end

	if gravity_x ~= 0 or gravity_y ~= 0 then
		Client.gravity = Vector2(TUNING.WORLD.GRAVITY * gravity_x, TUNING.WORLD.GRAVITY * gravity_y)
	end
end

Client.updateBalls = function (dt)
	-- Handle ball inputs
	if Client.controls.BallsAddBall.isDown then
		Balls:newRandomBall()
	end
	if Client.controls.BallsRemoveBall.isDown then
		if #Balls > 0 then
			Balls[#Balls] = nil
		end
	end
	if Client.controls.BallsClearAll.isDown and not Client.controls.BallsClearAll.wasDown then
		Balls:clear()
	end
	if Client.controls.BallsReset.isDown and not Client.controls.BallsReset.wasDown then
		Balls:reset()
	end

	--- Update existing Balls
	for i,Ball in ipairs(Balls) do
		if Client.gravity_enabled then
			Ball.velocity = Ball.velocity + Client.gravity
		end
		Ball.position = Ball.position + Ball.velocity*dt

		-- handle colisions with bounds
		if(Ball.position.x>Client.width-Ball.radius) then
			Ball.position.x = Client.width-Ball.radius
			Ball.velocity.x = -Ball.velocity.x
		elseif(Ball.position.x < 0+Ball.radius) then
			Ball.position.x = 0+Ball.radius
			Ball.velocity.x = -Ball.velocity.x
		end
		if(Ball.position.y>Client.height-Ball.radius) then
			Ball.position.y = Client.height-Ball.radius
			Ball.velocity.y = -Ball.velocity.y
		elseif(Ball.position.y < 0+Ball.radius) then
			Ball.position.y = 0+Ball.radius
			Ball.velocity.y = -Ball.velocity.y
		end
	end
end


return Client
